/*
 *	Character Set Recoding and Czech Support Module for Apache Web Server
 *
 *	(c) 1996-1998 Jakub Dadak, Intersoft s.r.o <web@intersoft.cz>
 *	(c) 1997-1998 Martin Mares <mj@atrey.karlin.mff.cuni.cz>
 *	(c) 1997-1998 Martin Plechsmid <plechsmi@karlin.mff.cuni.cz>
 *
 *	Home page: http://www.intersoft.cz/mod_czech/
 *
 *	This program can be freely used and distributed according to the terms
 *	of the GNU General Public License version 2 as present in file COPYING
 *	accompanying this package.
 */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <alloca.h>

#include "httpd.h"
#include "http_config.h"
#include "http_request.h"
#include "http_core.h"
#include "http_protocol.h"
#include "http_log.h"
#include "http_main.h"
#include "util_script.h"
#include "mod_alias.h"

#include "mod_czech.h"
#include "charconv/charconv.h"
#include "mod_czech.inc"

extern module alias_module;  /* Externi modul pro ziskani aliasu cgi */

/* Interni cisla charsetu */

#define CS_NONE		0	/* Bez konverze */
#define CS_ASCII	1
#define CS_ISO_8859_2	2
#define CS_WINDOWS_1250	3
#define CS_MAC_CE	4
#define CS_UTF_8	5
#define CS_KAMENICI	6
#define CS_KOI8		7
#define CS_CP_852	8
#define CS_ISO_8859_1	9
#define CS_MAX		10

/*
 * Zakladni tabulka charsetu prirazujici charsetum jejich standardni jmena,
 * kody pro konverzni knihovnu a priority pro charset negotiation (nejdrive
 * standardni kody, pak zbytek plnohodnotnych kody, nasledne aproximace
 * a nakonec ASCII).
 */

struct charset_info {
	int conv_code;		/* Poradi v charconv/chartable.in */
	char *mime_name;	/* Jmeno dle MIME */
	int nego_priority;	/* Priorita pro negotiation */
	char *comment;		/* Komentar pro chcp.html (v iso-8859-2) */
};

static struct charset_info charset_info[CS_MAX] = {
	{ -999,			"x-no-code",	0,	NULL 				},
	{ 0,			"US-ASCII",	1,	"ASCII (bez diakritiky)"	},
	{ 2,			"iso-8859-2",	10,	"ISO-8859-2 (ISO Latin-2)"	},
	{ 3,			"windows-1250",	5,	"Code Page 1250 (Windows-1250, WinCS, WinEE)" },
	{ 7,			"x-mac-ce",	4,	"Macintosh Central European (MacCE, CE, Mac)" },
	{ CONV_CHARSET_UTF8,	"UTF-8",	20,	"UniCode UTF-8"			},
	{ 4,			"x-kam-cs",	4,	"Brati Kamenit (KEYBCS2, CP859)" },
	{ 5,			"CSN_369103",	4,	"KOI-8 S2 (KOI8CS)"		},
	{ 6,			"cp852"	,	4,	"CP852 (PC Latin-2)"		},
	{ 1,			"iso-8859-1",	2,	"ISO-8859-1 (ISO Latin-1)"	},
};

/*
 * Debugging.
 */

#undef DEBUG

#ifdef DEBUG
#define DBG(x) printf x
#define DFLUSH fflush(stdout)
#else
#define DBG(x)
#define DFLUSH
#endif

/*
 * Zname tagy a jmena parametru, ktere jim menime.
 */

#define TAGS_NUM 11 
static struct {
        char tag_pref[10];
        char tag_url[10];
} tags[TAGS_NUM] = { {"a",           "href"},
                     {"link",        "href"},
                     {"base",        "href"},
                     {"frame",       "src" },
                     {"iframe",      "src" },
                     {"area",        "href"},
                     {"form",        "action"},
                     {"blockquote",  "cite"},
                     {"q",           "cite"},
                     {"ins",         "cite"},
                     {"del",         "cite"},
/* U nekterych tagu nema asi smysl prefix do URL pridavat:
                     {"image",   "src"},
                     {"input",   "src"},
                     {"body",    "background"}
                     {"applet",  "codebase"}
                     {"script",  "source"}
 * Nektere tagy mohou mit i nekolik URL parametru, ktere by mohly
 * prefix potrebovat. Nektere parametry mohou mit i dost slozitou
 * syntaxi (jsou zvyrazneny velkym pocatecnim pismenem):
                     {"object",  ("classid"),"codebase","data","Archive",("usemap")},
                     {"head",    "Profile"},
                     {"image",   ("src"),"longdesc",("usemap")},
                     {"input",   ("src"),("usemap")},
                     {"frame",   "src","longdesc"},
                     {"iframe",  "src","longdesc"},
 */
                   };

/*
 * Kontextovy buffer pro funkci EncodeStream().
 */

#define TAG_BUFF_SIZE 2000

struct tag_buff {
  char buf[TAG_BUFF_SIZE];		/* Buffer pro tagy */
  char *pos;				/* Aktualni pozice v bufferu */
  int  state;				/* Stav parseru HTML */
  int  interesting;			/* Prave zpracovavany tag je pro nas zajimavy */
};

/*
 * Pomocna makra pro konfiguraci modulu.
 */

#define UNDEFINED -1

#define mainCZ CZ_STATE(r->main)
#define prevCZ CZ_STATE(r->prev)

#define MERGE_INTS(old, new) ((new < 0) ? old : new)

/*
 * Modul osobne.
 */

module czech_module;

/*
 * Prototypy funkci.
 */

int IsContentTypeForEncoding(request_rec *r, char *content_type);
static int AddCharsetToMETA(request_rec *r, char *buff);
static int AddPrefixToHREF(request_rec *r, char *url);
static int IsDefaultPrefix(czech_server_config *sconf, char *prefx);
static char* GetDefaultPrefix(czech_server_config *sconf, int cpage);

/* ------------------------------------------------------------ */
/* --------------------- POMOCNE RUTINY ----------------------- */
/* ------------------------------------------------------------ */

/* ------------------- strcasestr ------------------------- */
/* Jako strstr(), ale case-insensitive.
 */
static char* strcasestr (char *str, char *pattn)
{
  register int i;
  char pattn0 = tolower(pattn[0]);

  for (; *str; str++)
  {
    if (tolower(*str) == pattn0)
    {
        for (i = 1; tolower(str[i]) == tolower(pattn[i]); i++);
	if (pattn[i] == '\0')
	    return (str);
    }
  }
  return NULL;
}

/* ----------------- TableGet ------------------- */
/* Totez jako table_get, ale case sensitive.
 */
static char *TableGet (const table *t, const char *key)
{
    table_entry *elts = (table_entry *)t->elts;
    int i;

    if (key == NULL) return NULL;
    
    for (i = 0; i < t->nelts; ++i)
        if (!strcmp (elts[i].key, key))
	    return elts[i].val;

    return NULL;
}

/* Totez jako table_set, ale case sensitive.
 */
static void TableSet (table *t, const char *key, const char *val)
{
  int i, j, k;
  table_entry *elts = (table_entry *)t->elts;
  int done = 0;

  for (i = 0; i < t->nelts; ++i)
    if (!strcmp (elts[i].key, key)) {
      if (!done) {
	elts[i].val = pstrdup(t->pool, val);
	done = 1;
      }
      else {     /* delete an extraneous element */
	for (j = i, k = i + 1; k < t->nelts; ++j, ++k) {
	  elts[j].key = elts[k].key;
	  elts[j].val = elts[k].val;
	}
	--t->nelts;
      }
    }

  if (!done) {
    elts = (table_entry *)push_array(t);
    elts->key = pstrdup (t->pool, key);
    elts->val = pstrdup (t->pool, val);
  }
}

/* ----------------- table_set_element ------------------- */
/* Jako table_set, ale value nemusi byt retezec (ale treba jina tabulka).
 * (Funkci table_get neni nutno upravovat, funguje dobre.)
 */

static void table_set_element (table *t, const char *key, char *val)
{
    register int i, j, k;
    table_entry *elts = (table_entry *)t->elts;
    int done = 0;

    for (i = 0; i < t->nelts; ++i)
	if (!strcasecmp (elts[i].key, key)) {
	    if (!done) {
	        elts[i].val = pstrdup(t->pool, val);
	        done = 1;
	    }
	    else {     /* delete an extraneous element */
                for (j = i, k = i + 1; k < t->nelts; ++j, ++k) {
                    elts[j].key = elts[k].key;
                    elts[j].val = elts[k].val;
                }
                --t->nelts;
	    }
	}

    if (!done) {
        elts = (table_entry *)push_array(t);
        elts->key = pstrdup (t->pool, key);
        elts->val = val;
    }
}

/*
 * Konverze kanonicke jmeno charsetu -> cislo charsetu.
 */
int find_charset(char *name)
{
  int i=0;

  for(i=0; i<CS_MAX; i++)
    if (!strcasecmp(charset_info[i].mime_name, name))
      return i;
  return -1;
}

/* --------------------- GetPrefixCode -------------------- */
static int GetPrefixCode(czech_server_config *sconf,char *cp)
{
  return find_charset(TableGet(sconf->czprefixes,cp));
}

/*
 * Konverze jmeno charsetu -> cislo charsetu pro konfig. soubory.
 */
static int config_charset(czech_server_config *sconf, char *name)
{
  int i;
  char *c;

  if ((i = find_charset(name)) >= 0)
    return i;
  else if ((c = table_get(sconf->charset_aliases, name)) &&
	   (i = find_charset(c)) >= 0)
    return i;
  return -1;
}

/*
 * Kanonicke jmeno charsetu.
 */
#define mime_charset(i) (charset_info[i].mime_name)

/* ------------------------------------------------------------ */
/* --------------------- KODOVANI BUFFERU --------------------- */
/* ------------------------------------------------------------ */

/*
 * Inicializace konverzniho mechanismu.
 */
static struct conv_context *init_conv_ctxt(request_rec *r, int from, int to)
{
  struct conv_context *cc;

  if (!(cc = CZ->conv_ctxt))
      cc = CZ->conv_ctxt = palloc(r->pool, sizeof(struct conv_context));
  conv_set_charset(cc,
		   charset_info[from].conv_code,
		   charset_info[to].conv_code);
  return cc;
}

/*
 * Prekodovani textu rovnou do vystupniho socketu.
 */
static int encode_and_send(request_rec *r, const char *first, int len)
{
  struct conv_context *c = CZ->conv_ctxt;
  char ccbuf[4096];
  int flag, n;

  c->source = first;
  c->source_end = first + len;
  c->dest = c->dest_start = ccbuf;
  c->dest_end = ccbuf + sizeof(ccbuf);
  do
    {
      flag = conv_run(c);
      if (flag & (CONV_DEST_END | CONV_SOURCE_END))
	{
	  char *ss = c->dest_start;
	  int sl = c->dest - c->dest_start;
	  while (sl)
	    {
	      if (r->connection->aborted)
		return -1;
	      n = bwrite(r->connection->client, ss, sl);
	      if (n < 0)
		return n;
	      sl -= n;
	      ss += n;
	    }
	  c->dest = c->dest_start = ccbuf;
	}
    }
  while (! (flag & CONV_SOURCE_END));
  return len;
}

/* -------------------  AttribValue    -----------------------*/
/* Hleda v retezci "attrib=value" a vrati ukazatel na value, nebo NULL.
 */
static char *AttribValue (register char *p, char *attrib)
{
  register int i;
  char attrib0 = attrib[0];

  for (; *p; p++)
  {
    if (tolower(*p) == attrib0)
        for (i = 1;; i++)
	{
	  if (tolower(p[i]) != attrib[i])
	      if (isalnum(attrib[i]))
		  break;                     /* slovo v bufferu p je delsi */
	      else
	      {
		  p += i;
		  while (*p  && isspace(*p))  p++;
		  if (*p++ != '=')  return NULL;
		  while (*p  && isspace(*p))  p++;
		  return (p);                                 /* navrat OK */
	      }
	}
    else if (*p == '"'  || *p == '\'')
        if (!(p = strchr(p + 1, *p)))
	    return NULL;
  }

  return NULL;
}

/* ------------------- GetValue ------------------------- */
/* Nalezeni hodnoty tagu: '="value"';  vraci pointer a v len delku.
 */
static char *GetValue(char *p, int *len)
{
  char *value;

  /* Napred preskocim mezery */
  while (*p  && isspace(*p)) p++;

  /* osetreni uvozovek */
  if (*p == '\''  || *p == '"')
  {
    value = p + 1;
    if ((p = strchr(p + 1, *p)) == NULL)
        return NULL;
  }
  else
  {
    value = p;
    while (*p  && !isspace(*p))  p++;

    /* Ted urcim konec nalezene "value" */
    if (*p == '\0'  && p[-1] == '>')
        if (p[-2] == '-'  && p[-3] == '-')             /* rovno "-->"? */
	  p -= 3;
	else
	  p--;
  }

  *len = p - value;
  return (value);
}

/* -------------- change_tag -------------- */
/* Doplni do prip. URL v HTML-tagu prefixy, zmeni v "MIME" charset,
 * nebo tag jinak upravi. Vraci, byl-li tag zmenen.
 */
static int change_tag(request_rec *r)
{
  int len, cpi, j;
  char *tagname, *tagvalue, *urlpos;
  char *buf = CZ->tag->buf;
  czech_dir_config *cfg = CZ_DIR_CONFIG(r);

#ifdef CZECH_HACKS
  if (cfg->allow_dyn_cp_changes && !strncmp(buf, "<!--", 4))
    {
      tagname = buf + 4;
      /* Odrezu mezery pred nazvem tagu */
      while (*tagname  && isspace(*tagname))  tagname++;

      /* Ted nactu jmeno tagu */
      for (len = 0; isalnum(tagname[len]); len++);

      /* Zmena source kodovani direktivou "<!--mycharset=xxx-->" : */
      if (!strncasecmp(tagname, "mycharset", len))
	{
	  tagvalue = tagname + len;
	  /* Preskocim mezery */
	  while (*tagvalue  && isspace(*tagvalue)) tagvalue++;

	  if (*tagvalue++ == '='  &&
	      (tagvalue = GetValue(tagvalue, &len)))
	    {
	      /* Rozlustim charset */
	      tagvalue[len] = '\0';      /* vse nasledujici ignoruji */

	      if ((cpi = config_charset(CZ_CONFIG(r), tagvalue)) >= 0)
		{
		  CZ->czsource = cpi;
		  CZ->master_cf->mycharset = cpi;
		  buf[0] = 0;
		  return 1;
		}
	      sprintf(buf, "[Error: unknown source charset]");
	      return 1;
	    }
	}
      return 0;
    }
#endif

  tagname = buf + 1;
  /* Odrezu mezery pred nazvem tagu */
  while (*tagname  && isspace(*tagname))  tagname++;

  /* Ted nactu jmeno tagu */
  for (len = 0; isalnum(tagname[len]); len++);

  if (isspace(tagname[len]))
    {
      /* Uprava charsetu v tagu META: */
      if (len == 4  && !strncasecmp(tagname, "meta", len))
	{
	  if ((tagvalue = AttribValue (tagname + len + 1, "http-equiv"))  &&
	      (tagvalue = GetValue(tagvalue, &len))  &&
	      !strncasecmp(tagvalue, "content-type", len))
	    {
	      if ((tagvalue = AttribValue (tagvalue + len, "content"))  ||
		  /* http-equiv a content mohou byt prohozeny */
		  (tagvalue = AttribValue (tagname + len + 1, "content")))
		{
		  cpi = AddCharsetToMETA(r, tagvalue);
		  if (cpi >= 0  && cfg->allow_dyn_cp_changes) {
		      if (cpi > 0)
			  CZ->czsource = cpi;
		      else {
			  CZ->czsource = CZ->czcode;     /* nekodovat */
		      }
		      init_conv_ctxt (r, CZ->czsource, CZ->czcode);
		  }
		  return 1;
		}
	    }
	}
      /* Pridani kodovaciho prefixu do direktiv obsahujicich URL: */
      else if (CZ->add_prefix)
	{
	  /* Rozpoznam direktivu a prip. dodam prefix do jejiho URL */
	  for (j = 0; j < TAGS_NUM; j++)
	    if (!strncasecmp(tagname, tags[j].tag_pref, len)  &&
		tags[j].tag_pref[len] == '\0')
	      {
	        /* Najdu prislusny atribut s URL a pridam prefix */
		if (urlpos = AttribValue (tagname + len + 1, tags[j].tag_url))
		  return AddPrefixToHREF(r, urlpos);
		break;
	      }
	}
    }

  return 0;
}

/* ------------- EncodeStream ------------ */

/*
 * "On the fly" prekodovavani streamu. Paklize se jedna o HTML, hledame
 * jednak zabudovane tagy pro zmenu kodovani, jednak take prepisujeme
 * URL v odkazech, obsahoval-li odkaz na puvodni dokument explicitni
 * prefix. Parsing HTML je resen jednoduchym konecnym automatem
 * se slozitym bufferingem nekompletnich tagu.
 */

enum html_states { S_HTML, S_LEFT, S_COMM, S_COMM1, S_COMM2, S_COMM_INSIDE,
	S_COMM_E1, S_COMM_E2, S_TAG, S_TAG1, S_TAG_Q, S_TAG_A };

#define SEND(r,b,e)				\
do if (b < e) {					\
     n = encode_and_send(r, b, e-b);		\
     if (n < 0) return n;			\
   } while(0)

#undef DEBUG_ENCODESTREAM		/* Ladeni EncodeStream na malych fragmentech */

#ifdef DEBUG_ENCODESTREAM
#define MAXX 10
int do_EncodeStream(request_rec *, const char *, int);
int EncodeStream(request_rec *r, const char *b, int l)
{
  int n, c, t = 0;
  while (l)
    {
      c = (l > MAXX) ? MAXX : l;
      n = do_EncodeStream(r, b, c);
      if (n < 0)
	return n;
      t += n;
      b += c;
      l -= c;
    }
  return t;
}
#else
#define do_EncodeStream EncodeStream
#endif

int do_EncodeStream(request_rec *r, const char *buff, int num)
{
  struct tag_buff *S = CZ->tag;		/* Stav parseru */
  const char *in_end = buff + num;	/* Konec vstupniho bufferu */
  const char *flush_first = buff;	/* Prvni dosud nezapsany byte */
  const char *flush_stop = in_end;	/* Posledni byte k zapsani */
  const char *c = buff;			/* Prave zpracovavany znak */
  unsigned int state = S->state;	/* Stav automatu */
  const char *tag_start = NULL;		/* Zacatek tagu ve vstupnim bufferu */
  char *tag_buf_pos = S->pos;		/* Konec tagu v tagovem bufferu */
  int interesting = S->interesting;	/* Mozna to je pro nas zajimavy tag */
  int len, n;
  char *d;

  if (tag_buf_pos)			/* Mame rozdelany tag? */
    tag_start = c;

  while (c < in_end)
    {
      switch (state)
	{
	default:			/* S_HTML */
	  if (!CZ->is_html)		/* Nutno testovat zde (shtml dokument muze dodavat cast vnitrku tagu) */
	    goto done;
	  while (c < in_end)
	    {
	      if (*c == '<')
		{
		  interesting = 1;
		  state = S_LEFT;
		  tag_start = c;
		  goto next;
		}
	      c++;
	    }
	  goto done;

	case S_LEFT:			/* Zacatek mozneho tagu */
	  if (*c == '!')
	    { state = S_COMM; break; }
	  if (isalpha(*c))
	    { state = S_TAG; break; }
	  goto mismatched_tag;

	case S_COMM:			/* Mozna to je komentar */
	  if (*c == '-')
	    state = S_COMM1;
	  else if (isalpha(*c))
	    state = S_TAG;
	  else
	    goto mismatched_tag;
	  break;
	case S_COMM1:			/* <!- */
	  if (*c != '-') goto mismatched_tag;
	  state = S_COMM2;
	  break;
	case S_COMM2:			/* <!-- */
#ifndef CZECH_HACKS
	  interesting = 0;
#endif
	  state = S_COMM_INSIDE;
	  /* fall-thru */
	case S_COMM_INSIDE:		/* Uvnitr komentare */
	  if (*c == '-') state = S_COMM_E1;
	  break;
	case S_COMM_E1:			/* - */
	  state = (*c == '-') ? S_COMM_E2 : S_COMM_INSIDE;
	  break;
	case S_COMM_E2:			/* -- */
	  if (*c == '>')		/* --> konec komentare (pozor, toto neni semantika dle RFC, */
	    goto tag_end;		/* ale ta, na ktere se shoduje vetsina browseru :() */
	  if (*c != '-')
	    state = S_COMM_INSIDE;
	  break;

	case S_TAG:			/* Tag */
	  if (*c == ' ' || *c == '\t' || *c == '\n')
	    state = S_TAG1;
	  else if (*c == '>')
	    goto tag_end;
	  break;
	case S_TAG1:			/* Parametry tagu */
	  if (*c == '>')
	    goto tag_end;
	  else if (*c == '"')
	    state = S_TAG_Q;
	  else if (*c == '\'')
	    state = S_TAG_A;
	  break;
	case S_TAG_Q:			/* Vyraz v uvozovkach */
	  if (*c == '"')
	    state = S_TAG1;
	  else if (*c == '\n')
	    goto mismatched_tag;
	  break;
	case S_TAG_A:			/* Vyraz v apostrofech */
	  if (*c == '\'')
	    state = S_TAG1;
	  else if (*c == '\n')
	    goto mismatched_tag;
	  break;

	tag_end:
	  len = ++c - tag_start;
	  if (!tag_buf_pos)
	    tag_buf_pos = S->buf;
	  if (interesting && tag_buf_pos + len < S->buf + TAG_BUFF_SIZE - 100)
	    {
	      memcpy(tag_buf_pos, tag_start, len); /* Tag, ktery mozna budeme menit */
	      tag_buf_pos[len] = 0;
	      if (change_tag(r))
		{			/* A opravdu menime */
		  SEND(r, flush_first, tag_start);
		  d = S->buf + strlen(S->buf);
		  if (d > S->buf)
		    SEND(r, S->buf, d);
#ifdef CZECH_HACKS
		  else			/* change_tag() asi zmenila czsource */
		    init_conv_ctxt (r, CZ->czsource, CZ->czcode);
#endif
		  flush_first = c;
		  tag_start = tag_buf_pos = NULL;
		  state = S_HTML;
		  goto reswitch;
		}
	    }
	mismatched_tag:			/* Tag je nekompletni nebo jej nemenime */
	  if (tag_buf_pos)		/* Takze pouze vylijeme pripadna nabufferovana data */
	    SEND(r, S->buf, tag_buf_pos);
	  tag_start = tag_buf_pos = NULL;
	  state = S_HTML;
	  goto reswitch;
	}
    next:
      c++;
    reswitch:
    }

done:
  if (tag_start)			/* Mame rozdelany tag */
    {
      len = c - tag_start;
      if (!tag_buf_pos)
	tag_buf_pos = S->buf;
      if (interesting && tag_buf_pos + len < S->buf + TAG_BUFF_SIZE - 100)
	{
	  memcpy(tag_buf_pos, tag_start, len);
	  tag_buf_pos += len;
	  flush_stop = tag_start;
	}
      else				/* Nevejdeme se nebo je to nezajimavy tag */
	{
	  SEND(r, S->buf, tag_buf_pos);
	  interesting = 0;
	  tag_buf_pos = NULL;
	}
    }

  SEND(r, flush_first, flush_stop);
  S->pos = tag_buf_pos;
  S->state = state;
  S->interesting = interesting;
  return num;
}

#undef SEND

/* -------------------- FlushModCzechBuffer -------------- */
static int FlushModCzechBuffer (request_rec *r)
{
  struct tag_buff *S = CZ->tag;
  int n = 0;

  /* Je-li nejaky nekompletni tag, vysypmez jej do vystupu... */
  if (S && S->pos)
    {
      n = encode_and_send(r, S->buf, S->pos - S->buf);
      S->pos = NULL;
    }
  return n;
}


/* ------------------------------------------------------------ */
/* --------------------- PRIDAVANI PREFIXU -------------------- */
/* ------------------------------------------------------------ */

/* -------------------- IsModCzechServer -------------- */
/* Zjisti, zda server "name:port" podporuje ceske prefixy.
 */
static int IsModCzechServer (request_rec *r, char *name, int port)
{
  server_rec *server = r->connection->base_server;
  czech_dir_config *cfg = CZ_DIR_CONFIG(r);
  array_header *srvports;
  int i;

  /* Nejprve projedu moje virtualni servery */
  while (server) {
      if (!strcasecmp(name, server->server_hostname) &&
	  (server->port == port))
      {
	/* Overim, ze ma zapnuty modul */
	czech_dir_config *cc = get_module_config(server->lookup_defaults, &czech_module);
	if (cc->encode)
	  return 1;
      }
      server = server->next;
  }

  /* Ted projedu "bratrske mod_czech servery" */
  if (srvports = (array_header*) table_get(cfg->czech_servers, name)) {
      if (!srvports->nelts)
          return (1);
      else
      {
          for (i = 0; i < srvports->nelts; i++) {
	      if ((int) srvports->elts[i] == port)  return (1);
	  }
      }
  }

  return (0);
}

/* -------------------- PathIfAddPrefix ------------ */
/* Vrati ukazatel na misto, kam se ma vlozit prefix (pokud se ma vlozit).
 */
static char *PathIfAddPrefix (request_rec *r, char *url)
{
  char *p, *p1, c, c1;
  int port;

  if (!CZ->add_prefix)  return NULL;

  /* skip spaces */
  for (; *url && isspace(*url); url++);

  if (*url == '"') {
      c = '"';
      url++;
  }
  else
  if (*url == '\'') {
      c = '\'';
      url++;
  }
  else
      c = '>';                /* end of HTML-tag */

  if (*url != '/')
    if (!strncasecmp(url, "http:/", 6))
    {
      p = url + 5;
      if (url[6] == '/')  {
	/* server name is in url */
	for (p += 2; *p; p++)
        {
	  if (*p == ':' || *p == '/' || *p == ';' || *p == '?' ||
	      *p == c  || isspace(*p))
	    break;
	}
	p1 = p;
	if (*p == ':') {      /* Port*/
	  p++;
	  if (!isdigit(*p))
	      port = DEFAULT_PORT;
	  else
	      port = (int) strtol(p, &p, 10);
	}
	else
	  port = DEFAULT_PORT;
	c1 = *p1;
	*p1 = '\0';           /* terminate server-name */
	if (!IsModCzechServer(r, url + 7, port)) {
	    *p1 = c1;
	    return NULL;
	}
	*p1 = c1;
	if (*p  &&  *p != '/'  && *p != ';'  && *p != '?'  &&
	    *p != c  && !isspace(*p))
	    return NULL;
      }
    }
    else
      return NULL;            /* return if neither http: nor absolute path */
  else
    p = url;

  if (*p == '/') {            /* look for (default) prefix in abs. path */
      for (p1 = p + 1; *p1; p1++)
	if (*p1 == '/'  || *p == ';'  || *p1 == '?'  ||
	    *p1 == c  || isspace(*p1))
	    break;
      c1 = *p1;
      *p1 = '\0';
      if (IsDefaultPrefix(CZ_CONFIG(r), p + 1)) {
	  *p1 = c1;
	  return NULL;
      }
      *p1 = c1;
  }

  return p;
}

/* -------------------- FindPrefix ------------ */
/* Najde v ceste prefix, pokud tam je, a vrati jej (z cesty jej nevynda).
 */
static char *FindPrefix (request_rec *r, char *path)
{
  czech_server_config *sconf = CZ_CONFIG(r);
  table_entry *elts = (table_entry *) sconf->czprefixes->elts;
  int nelts = sconf->czprefixes->nelts;
  char *prefx, c;
  int i, len;

  if (*path != '/')  return NULL;
  path++;
  for (i = 0; i < nelts; i++) {
      prefx = elts[i].key;
      len = strlen(prefx);
      if (!strncmp(path, prefx, len)) {
	  c = path[len];
	  if ( c == '/'  || c == ';'  || c == '?'  || !c  || isspace(c)  ||
	       c == '"'  || c == '\''  || c == '>')
	      return (prefx);
      }
  }

  return NULL;
}


/* -------------------- AddPrefixToURL ------------ */
/* Prida prefix do URL, pokud tam patri.
 */
char *AddPrefixToURL (request_rec *r, char *url)
{
  czech_server_config *sconf = CZ_CONFIG(r);
  char *abspath, *srvend, *prefx, c[2], c1;

  if ((abspath = PathIfAddPrefix(r, url))) {
    srvend = abspath;
    if (prefx = FindPrefix(r, abspath))
        abspath += 1 + strlen(prefx);        /* prefix begins with a slash */
    c[1] = '\0';                  /* end of string */
    if (*c = *abspath)
        abspath++;
    c1 = *srvend;
    *srvend = '\0';               /* bylo to '/' nebo '\0' ap., ozelime je */
    url = pstrcat (r->pool, url, "/", GetDefaultPrefix(sconf,
                       (prefx ? GetPrefixCode(sconf,prefx) : CZ->czcode)),
		   "/", (*c != '/' ? c : ""), abspath, NULL);
    *srvend = c1;                 /* vratit lomitko nebo '\0' zpet */
  }
  return url;
}

/* -------------------- AddPrefixToHREF  --------------------------*/
/* Totez jako AddPrefixToURL, ale v bufferu.
 * Konec textu v bufferu musi byt vyznacen '\0'!
 */
static int AddPrefixToHREF(request_rec *r, char *url)
{
  czech_server_config *sconf = CZ_CONFIG(r);
  char *abspath, *srvend, *prefx;
  int len;

  if ((abspath = PathIfAddPrefix(r, url))) {
    srvend = abspath;
    if (prefx = FindPrefix(r, abspath))
        abspath += 1 + strlen(prefx);        /* prefix begins with a slash */
    prefx = GetDefaultPrefix(sconf,
		     (prefx ? GetPrefixCode(sconf,prefx) : CZ->czcode));
    len = strlen(prefx);
    memmove(srvend + 1 + len, abspath, strlen(abspath) + 1);
    *srvend = '/';
    memcpy(srvend + 1, prefx, len);
    return 1;
  }
  else
    return 0;
}

/* ------------------------------------------------------------ */
/* --------------------- PRIDAVANI PREFIXU -------------------- */
/* ------------------------------------------------------------ */

/* -------------------- AddCharsetToMETA  --------------------------*/
/* Do hodnoty Content-Type doplni (resp. upravi) charset podle czcode.
 * Vrati puvodni hodnotu charsetu, nebo -1.
 * Konec textu v bufferu musi byt vyznacen '\0'!
 */
static int AddCharsetToMETA(request_rec *r, char *buff)
{
  char *strend, *pos, *pb, *pe, c, c1, s[100];
  int len, len2, cpi = -1;

  if ((c = *buff) == '"'  || c == '\'')
  {
      buff++;
      if (strend = strchr(buff, c))
	  *strend = '\0';
      else
	  return (-1);
  }
  else
      return (-1);              /* jinak bychom tam museli dat ty uvozovky */

  if (IsContentTypeForEncoding(r, buff))
  {
      pb = buff;
      while (pos = strchr(pb, ';'))
      {
	  for (pb = pos; *++pb == ' '; );
	  if (!strncasecmp(pb, "charset=", 8))
	  {
	      if (CZ_DIR_CONFIG(r)->allow_dyn_cp_changes) {  /* rozlustim puvodni charset */
		  pe = pb += 8;
		  while (*pe && *pe != ',' && *pe != ';' && *pe != ' ')
		      pe++;
		  c1 = *pe;
		  *pe = '\0';
		  cpi = config_charset(CZ_CONFIG(r), pb); /* povoleny jsou i aliasy charsetu */
		  *pe = c1;
	      }
	      len = pe - pos;
	      break;
	  }
      }
      if (!pos)
      {
	  for (pos = pb; *pos; pos++);
	  len = 0;
      }

      if (c)  *strend = c;

      if (cpi  && CZ->czcode != CS_ASCII  && CZ->czcode != CS_ISO_8859_1)
	  sprintf(s, "; charset=%s", mime_charset(CZ->czcode));
      else
	  *s = '\0';
      len2 = strlen(s);

      memmove(pos + len2, pos + len, strlen(pos + len)+1);
      memcpy(pos, s, len2);
      return cpi;
  }

  if (c)  *strend = c;
  return -1;
}

/* ------------------------------------------------------------ */
/* ------------------------ ANALYZA URL ----------------------- */
/* ------------------------------------------------------------ */

/*---------------- GetDefaultPrefix --------------- */
/* Vrati defaultni prefix prislusny cpage.
 */
static char* GetDefaultPrefix(czech_server_config *sconf,int cpage)
{
  table_entry *elts = (table_entry*)sconf->czprefixes->elts;
  int         nelts = sconf->czprefixes->nelts;
  char *cp = (cpage < 0) ? mime_charset(0) : mime_charset(cpage);
  int i;

  for(i=0; i < nelts; i++)  
  {
    if (!strcmp(cp,elts[i].val))
      return elts[i].key;
  }
  return NULL;
}

/*---------------- IsDefaultPrefix --------------- */
/* Zjisti, zda je dodany retezec standardnim prefixem.
 * Standartni je prvni prefix v tabulce odpovidajici prislusne CP.
 */
static int IsDefaultPrefix(czech_server_config *sconf,char *prefx)
{
  char *deflt,*cp;

  if ((cp = TableGet(sconf->czprefixes, prefx)) &&
      (deflt = GetDefaultPrefix(sconf,find_charset(cp))) &&
      !strcmp(prefx,deflt))
    return 1;
  else
    return 0;
}

/* ----------------  IsPostfxStr ------------- */
/* Fce otestuje, jestli je zacatek str postfix, jestli ano, tak
 * ten postfix priradi do *postfx a prislusnou CP priradi do *cpage.
 * Jinak cp nezmeni.
 */
static int IsPostfxStr(czech_server_config *sconf,char *str,char **postfx, 
		       char **cpage)
{
table_entry *elts = (table_entry*)sconf->postfix_aliases->elts;
int         nelts = sconf->postfix_aliases->nelts;
char *cp;
int i, len;
 
  for(i=0; i < nelts; i++)
    {
      len = strlen(elts[i].key);
      if (!strncmp(str, elts[i].key, len) &&
	  (!str[len] || str[len] == '/') &&
	  (cp = TableGet(sconf->czprefixes, elts[i].val)))
	{
	  *postfx = elts[i].key;
	  *cpage = cp;
	  return 1;
	}
    }
  return 0;
}

/* -------------------- IsCgiRequest  --------------------------*/
/* Zjisti, zda je aktualni request volanim cgi-skriptu.
 * Jestli ano, vrati pozici, odkud se ma hledat pripadny postfix.
 */
static int IsCgiRequest(request_rec *r, char *request)
{
 void *sconf;
 alias_server_conf *serverconf;
 int i;
 alias_entry *entries;
 alias_entry *p;

 sconf = r->server->module_config;
 serverconf = (alias_server_conf *)get_module_config(sconf, &alias_module);
 entries = (alias_entry *)serverconf->aliases->elts;

 if(*request != '/') /* neni zadan href uz od rootu */
 {
  for(i=0; i< serverconf->aliases->nelts; ++i)
  {
   p = &entries[i];
   if(!strncmp(p->fake, r->uri, strlen(p->fake)) &&
      p->handler && !strcmp(p->handler,"cgi-script")) 
     return(1);
  }
 }
 else
 for(i=0; i< serverconf->aliases->nelts; ++i)
 {
  p = &entries[i];
  if(!strncmp(p->fake, request, strlen(p->fake)) &&
     p->handler && !strcmp(p->handler,"cgi-script")) 
    return(strlen(p->fake));
 }
 return(0);
}

/* --------------- ChangeCGIRequest ------------- */
/* Fce vypreparuje CPxxx z requestu ten vrati zpet do
 * request a do *cp da prislusnou CP, vraci 1 kdyz to udelal,
 * jinak 0 a *cp nezmeni
 */
static int ChangeCGIRequest(czech_server_config *sconf,char *request,
			    int pos, char **cp)
{
int i,j;
char *postfx;
char virt_req[2000];
int in_fname=1;             /* ze rozebiram filename */
int czpage=0;

 strncpy(virt_req, request, pos);

 for(i=pos, j=pos; request[i]; i++)
 {
  if(in_fname)
  {
   if(request[i] == '/'  || request[i] == ';'  || request[i] == '?')
       in_fname=0;
 /* MARTIN: upravit (lepe) */
    if(i>0 && request[i-1] == '.')
    {
      if(IsPostfxStr(sconf,&request[i],&postfx,&*cp))
      { i+=strlen(postfx); j--; czpage=1;}
    }
   }
   virt_req[j++] = request[i];
  }
  virt_req[j]=0;
  if(czpage) strcpy(request,virt_req);

  return(czpage);
}


/* --------------- ChangeRequestPrefix ------------- */
/* Fce testuje normalni request na existenci CP zadane jako 
 * prefix v URI, request je modifikovan a CP kopirovana na cp. 
 *      Je- li CP defaultni, je vracena 1, 
 *      pokud neni defaultni, je vracena 2, 
 *      neni-li prefix nalezen, tak se vraci' 0.
 */
static int ChangeRequestPrefix(czech_server_config *sconf,char *request, char **cp)
 {
   table_entry *elts = (table_entry*)sconf->czprefixes->elts;
   int         nelts = sconf->czprefixes->nelts;
   char *p;
   int i,del;

  /* musi zacinat lomitkem */
   if (request[0] != '/')  return 0; 
   p = &request[1];

   for (i=0; i < nelts; i++)
   {
     if (!strncmp(p,elts[i].key, del=strlen(elts[i].key)) &&
	 (!p[del] || p[del] == '/'))
     {
       *cp = elts[i].val;
       strcpy(request,p+del);
       if (!*request)
	 {
	   *request++ = '/';
	   *request = 0;
	 }
       if (IsDefaultPrefix(sconf,elts[i].key))
	 return 1;
       else
	 return 2;
     }
   }
   return 0;
 }

/* ----------------  ChangeRequest -------------- */
/* Fce testuje normalni request na existenci CP, request
 * je modifikovan a CP kopirovana na cp a vracena 1 pri uspechu
 * jinak 0.
 */
static int ChangeRequest(czech_server_config *sconf,char *request, char **cp)
{
  char *dummy, *p;

  p = (p=strrchr(request,'/')) ? p : request;  /* Najdu si posledni '/' */
  if (!(p=strrchr(p,'.')))                     /* Najdu si posledni '.' */
    return 0;

  if(IsPostfxStr(sconf,p+1,&dummy,&*cp))
  {
    *p = '\0';
    return 1;
  }
  else
    return 0;
}

/* --------------- Analyze request --------------- */
/* Tato fce provede jako prvni na serveru analyzu
 * request_line a v pripade, ze objevi postfix nebo nestandardni
 * prefix pro urceni ceske kodove stranky , tak ho odstrani a 
 * udela redirect s ekvivalentnim prefixem; pokud prijde standardni
 * prefix, tak ho urizne a da do req->czcharset.
 * Postfix ma vetsi prioritu, nez prefix, takze pokud
 * prijde prefix i postfix, redirect jde s hodnotou 
 * postfixu 
 * Fce Vraci 0 = neobsazen prefix, ani postfix
 *           1 = obsazen prefix, explicitni kodovani
 *           2 = obsazen postfix nebo nestandardni prefix -> redirect
 */
static int AnalyzeRequest(request_rec *r, char *uri, int change_semafor)
{
 char request[2000]={0};
 char *cp;         
 int pos;
 int czrequest=0;
 int make_redir=0;
 czech_server_config *sconf = CZ_CONFIG(r);

 if(strlen(uri) < 1990)   strcpy(request,uri);
 else                     return(-1); 

 /* ........ nyni zacnu probirat request .............*/
 switch (ChangeRequestPrefix(sconf,request,&cp))
 {
  case 0: /* zadny prefix nenalezen */
          break;
  case 1: /* standardni (default) prefix */
          czrequest=1;
	  break;
  case 2: /* nestandardni prefix */
	  make_redir=1;
	  break;
 }

 /* ......... zjistim zda se nejedna o cgi request ...*/
 if(pos=IsCgiRequest(r,request))
 {
   /* ... odstranim CP z CGI requestu, ktery je cppostfix  */
   if(ChangeCGIRequest(sconf,request,pos,&cp)) make_redir=1;
 }
 else
 {
   /* .. odstranim CP z requestu , ktery je v cppostfix */ 
   if(ChangeRequest(sconf,request,&cp)) make_redir=1;
 }

 /* Nyni udelam redirect na URL ve forme prefixu (prisel postfix) */
 if(make_redir) 
 {
	strcpy(uri,request);
	CZ->czcode=find_charset(cp);
	if(CZ->czcode < 0) CZ->czcode = 0;
 	return(2); 
 }
 
 /* ........ nyni slozim zpatky request_line a nastavim potrebne..*/
 /* ........ jestlize je nastaveno czrequest, vim ze je explicitne
    zadana stranka */

 else if(czrequest)
 {
   strcpy(uri,request);
   if(change_semafor)
   {
	CZ->czcode=find_charset(cp); 
	if(CZ->czcode < 0) CZ->czcode = 0;
   }
   return(1);
 }
 else
   return(0);
}


/* ------------------------------------------------------------ */
/* ------------------- AUTODETEKCE KODOVANI ------------------- */
/* ------------------------------------------------------------ */

/*
 * Hledani v blacklistu.
 */
static int is_forbidden(char *val, char *list)
{
  char *z;
  int l = strlen(val);

  while (list && *list)
    {
      for(z=list; *z && *z != ','; z++);
      if (z - list == l && !strncasecmp(list, val, l))
	return 1;
      if (*z)
	z++;
      list = z;
    }
  return 0;
}

/*
 * Analyza Accept-Charsets. Ridi se q-faktory, ktere zada client,
 * v pripade shody pak nasimi vlastnimi prioritami.
 */
static int analyze_accept_charset(char *str, char *forbidden)
{
  int best=-1, bestq=-1, bestp=-1, q, c, code, dc;
  char *name, *nend;

  for(;;)
    {
      while (*str == ' ')
	str++;
      if (!*str)
	break;
      name = str;
      while (*str && *str != ',' && *str != ';')
	str++;
      nend = str;
      q = 1000;
      /*
       * Trocha cerne magie: q-faktor je budto 1 (s pripadnymi nulami za desetinnou
       * teckou) nebo 0 nebo 0. a max 3 desetinna mista za teckou. [RFC 2086]
       */
      if (!strncasecmp(str, ";q=0", 4))
	{
	  str += 3;
	  if (*str == '0')
	    {
	      q = 0;
	      if (str[1] == '.')
		{
		  str += 2;
		  dc = 0;
		  while (*str >= '0' && *str <= '9')
		    {
		      dc++;
		      q = 10*q + *str++ - '0';
		    }
		  while (dc++ < 3)
		    q *= 10;
		}
	    }
	}
      if (*str && *str != ',')
	str++;
      if (*str)
	str++;
      c = *nend;
      *nend = 0;
      code = (is_forbidden(name, forbidden)) ? -1 : find_charset(name);
      *nend = c;
      if (code >= 0)
	{
	  if (q > bestq)
	    {
	      bestq = q;
	      bestp = -1;
	    }
	  if (q == bestq && charset_info[code].nego_priority > bestp)
	    {
	      best = code;
	      bestp = charset_info[code].nego_priority;
	    }
	}
    }
  return best;
}

/*
 * Vesteni charsetu podle ruznych indicii -- nejprve podle headeru
 * Accept-Charset, neni-li pritomen, pak heuristicky podle jmena
 * clienta, v nejhorsim pripade fallback.
 */
static int guess_charset(request_rec *r)
{
  czech_dir_config *cfg = CZ_DIR_CONFIG(r);
  char *accept, *cs;
  int code;

  DBG(("Analyzing %s\n", r->uri));
  DBG(("Br=%s\n", table_get(r->headers_in, "User-Agent")));
  if (accept = table_get(r->headers_in, "Accept-Charset"))
    {
      DBG(("AC=%s\n", accept));
      cs = table_get(r->subprocess_env, "ignore_charset");
      DBG(("ignore=%s\n", cs));
      code = analyze_accept_charset(accept, cs);
      if (code >= 0)
	return code;
   }

  if (cfg->guess_charset &&
      (cs = table_get(r->subprocess_env, "browser_charset")) &&
      (code = analyze_accept_charset(cs, NULL)) >= 0)
	{
	  DBG(("bc=%s\n", cs));
	  return code;
	}

  return -1;
}

/*
 * Vesteni charsetu pro POST. Nejdrive se podivame, je-li v POST
 * requestu uveden primo charset, jinak pouzijeme default. Specifikaci
 * charsetu ovsem z content-typu odstranime, bude pozdeji nahrazena
 * spravnou.
 */
static int guess_post_charset(request_rec *r, char *ct)
{
  char *s, *t;
  int c, a;

  if (CZ->form_source != UNDEFINED)	/* Jiz uhodnuto drive */
    return CZ->form_source;

  while (ct = strchr(ct, ';'))
    {
      s = ct++;
      while (*ct == ' ')
	ct++;
      if (!strncasecmp(ct, "charset=", 8))
	{
	  t = ct += 8;
	  while (*t && *t != ',' && *t != ';' && *t != ' ')
	    t++;
	  c = *t;
	  *t = 0;
	  a = find_charset(ct);
	  *t = c;
	  strcpy(s, t);
	  return a;
	}
    }
  /*
   * Zde by samozrejme bylo korektni rici, ze CS_ASCII, ale mnozi clienti
   * nejsou schopni korektne Content-type uvest. Grr.
   */
  return CZ->czcode;
}

/* ------------------------------------------------------------ */
/* -------------------- ZAPINANI KODOVANI --------------------- */
/* ------------------------------------------------------------ */

/* -------------------- IsContentTypeForEncoding -------------- */
/* Fce vraci 1 = ano tento CT prekodujeme
             0 = ne nepodporovany CT        */
int IsContentTypeForEncoding(request_rec *r, char *content_type)
{
  czech_dir_config *cfg = CZ_DIR_CONFIG(r);
  char *p;

  if (!content_type)  return (0);

  if (p = strchr(content_type, ';'))
      *p = '\0';
  if (table_get(cfg->content_types, content_type)) {
      if (p)  *p = ';';
      return (1);
  }
  else {
      if (p)  *p = ';';
      return (0);
  }
}

/* -------------------- DecideWhetherEncode ------------ */
int DecideWhetherEncode (request_rec *r)
{
  czech_dir_config *cfg = CZ_DIR_CONFIG(r);

  DBG(("========== DecideWhetherEncode =========\n"));
  if (!IsContentTypeForEncoding(r, r->content_type)  ||
      r->content_encoding)
  {
       /* Tag->buff je treba flushnout, protoze ted se EncodeStream() 
	* nebude pouzivat. Jestli se bude kodovat, tak neflushovat!
	* Pripadna chyba pri flushovani se ignoruje, ale nemelo by to
	* vadit, protoze bwrite si s tim umi poradit.
	*/
      FlushModCzechBuffer(r);
      CZ->czcode = 0;
  }
  else if (!strncasecmp(r->content_type, "text/html", 9) && cfg->allow_html_parsing)
      CZ->is_html = 1;

  if (CZ->czcode > 0)
      init_conv_ctxt (r, CZ->czsource, CZ->czcode);
  DBG(("content-type='%s', czsource=%d, czcode=%d\n", r->content_type, CZ->czsource, CZ->czcode));
  return (CZ->czcode);
}

/* -------------------- SetupModCzechEncoding ------------ */
/* Tato fce nastavuje potrebne polozky v request_rec pro kodovani cestiny*/
void SetupModCzechEncoding (request_rec *r)
{
  char *url;
  czech_dir_config *cfg = CZ_DIR_CONFIG(r);

  if (!cfg->encode)  return;

  if (!CZ)
    CZ = pcalloc(r->pool, sizeof(mcz_config));

  DBG(("========== SetupModCzechEncoding =========\n"));
  DBG(("add_prefix=%d\n", CZ->add_prefix));
  if (CZ->add_prefix)  {
    /* Zde mohou vzniknout problemy pri zmene kodovani, pokud je provedena
     * redirekce na *relativni* URL. To ale nevadi, Apache totiz relativni
     * redirekce neumoznuje!
     */
    if ((url = table_get(r->headers_out, "Location")))
      table_set(r->headers_out, "Location", AddPrefixToURL(r, url));
    if ((url = table_get(r->headers_out, "Content-Location")))
      table_set(r->headers_out, "Content-Location", AddPrefixToURL(r, url));
    if ((url = table_get(r->headers_out, "Content-Base")))
      table_set(r->headers_out, "Content-Base", AddPrefixToURL(r, url));
  }

  DecideWhetherEncode(r);

  if (CZ->czcode != 0)
  {
    /*
     * Pokud je v URI kodovani explicitne, lze cacheovat. Pokud jsme se ovsem
     * rozhodli na zaklade Accept-Charset nebo User-Agent, musime budto cachi
     * rici, ze zalezi pri jinych hodnotach techto headeru nesmi nynejsi
     * odpoved pouzit (Vary u HTTP/1.1), nebo ji nesmi cacheovat vubec
     * (starsi protokoly).
     */
    if (!CZ->add_prefix) {
      table_merge(r->headers_out, "Vary", "User-Agent");
      table_merge(r->headers_out, "Vary", "Accept-Charset");
      if (!cfg->do_cache  &&  r->proto_num < 1001)
	table_merge (r->headers_out, "Pragma", "no-cache");
    }

    if(cfg->add_cp_to_ct  &&
       !table_get(r->subprocess_env, "no_ct_charset") &&
       CZ->czcode != CS_ASCII && CZ->czcode != CS_ISO_8859_1)
      {
	r->content_type = pstrcat(r->pool, r->content_type,
				  "; charset=", mime_charset(CZ->czcode),NULL);
      }
  }
}


/* ------------------------------------------------------------ */
/* ------------------- OSETRENI CGI-SKRIPTU ------------------- */
/* ------------------------------------------------------------ */

/*
 * Triky s hexadecimalnimi cisly.
 */
static int enhex(int x)
{
  x &= 0x0f;
  return '0' + x + ((x >= 10) ? 7 : 0);
}

static int dehex(int x)
{
  if (x >= '0' && x <= '9')
    return x - '0';
  x &= 0xdf;
  if (x >= 'A' && x <= 'F')
    return x - 'A' + 10;
  return 0;
}

struct filter {
  struct filter *next, *prev;
  request_rec *request;
  int finished;
  char *buf, *buf_end, *buf_pos, *buf_mark;
  int (*handler)(struct filter *f, char **src, char *se, char **dest, char *de);
};

#define F_IBE 1				/* Input buffer empty */
#define F_OBF 2				/* Output buffer full */
#define F_ERR 4

static void set_filter_buffer(struct filter *f, char *buf, int size)
{
  f->buf = f->buf_pos = f->buf_mark = buf;
  f->buf_end = buf + size;
}

static struct filter *
add_filter(request_rec *r,
	   int (*handler)(struct filter *f, char **src, char *se, char **dest, char *de),
	   unsigned bufsize,
	   unsigned structsize)
{
  struct filter *f = pcalloc(r->pool, structsize ? structsize : sizeof(struct filter));
  mcz_config *c = CZ;

  if (bufsize)
    set_filter_buffer(f, palloc(r->pool, bufsize), bufsize);
  f->handler = handler;
  f->request = r;
  if (c->last_filter)
    {
      c->last_filter->next = f;
      f->prev = c->last_filter;
      c->last_filter = f;
    }
  else
    c->first_filter = c->last_filter = f;
  return f;
}

static int run_filter_chain(request_rec *r)
{
  mcz_config *c = CZ;
  struct filter *f = c->last_filter;

  while (f)
    {
      int flags;
      struct filter *p = f->prev;
      struct filter *n = f->next;

      flags = f->handler(f,
			 &f->buf_pos,
			 f->buf_mark,
			 (n ? &n->buf_mark : NULL),
			 (n ? n->buf_end : NULL));
      if (flags & F_ERR)
	return -1;
      if (flags & F_OBF)
	f = f->next;
      else if (flags & F_IBE)
	{
	  if (!p || p->finished)
	    {
	      f->finished = 1;
	      f = f->next;
	    }
	  else
	    {
	      f->buf_pos = f->buf_mark = f->buf;
	      f = p;
	    }
	}
    }
  return 0;
}

static void reset_filters(request_rec *r)
{
  CZ->first_filter = CZ->last_filter = NULL;
}

/*
 * Filter handler pro zakodovani URL-encoded stringu.
 */
struct enescape_filter {
  struct filter f;
  int backslash_flag;
};

static int filter_enescape(struct filter *f, char **src, char *se, char **dst, char *de)
{
  struct enescape_filter *F = (void *) f;
  char *s = *src;
  char *d = *dst;
  int backslash = F->backslash_flag;

  if (backslash)
    goto backsl;
  while (s < se)
    {
      unsigned int c;

      if (d >= de)
	break;
      c = *s++;
      if (c >= 'A' && c <= 'Z' ||	/* Tyto znaky smeji byt kodovany primo -- viz RFC 1738 a 1808 */
	  c >= 'a' && c <= 'z' ||
	  c >= '0' && c <= '9' ||
	  c == '$' || c == '-' || c == '_' ||
	  c == '.' || c == '+' || c == '!' ||
	  c == '*' || c == '\'' || c == '(' ||
	  c == ')' || c == ',')
	*d++ = c;
      else if (c == ' ')
	*d++ = '+';
      else if (c == '\\')		/* Specialni znaky -- nikdy nekodovat */
	{
	  backslash = 1;
	backsl:
	  if (s >= se)
	    break;
	  *d++ = *s++;
	  backslash = 0;
	}
      else
	{
	  if (d + 3 > de)
	    break;
	  *d++ = '%';
	  *d++ = enhex(c >> 4);
	  *d++ = enhex(c);
	}
    }
  *src = s;
  *dst = d;
  F->backslash_flag = backslash;
  if (s >= se)
    return F_IBE;
  return F_OBF;
}

/*
 * Filter handler pro dekodovani URL-encoded stringu.
 */
struct deescape_filter {
  struct filter f;
  unsigned char state[4];
};

static int filter_deescape(struct filter *f, char **src, char *se, char **dst, char *de)
{
  struct deescape_filter *F = (void *) f;
  char *xbuf = F->state;
  int state = strlen(xbuf);
  char *dbp = *dst;
  char *from = *src;

  if (state)
    goto percent;
  while (from < se)
    {
      unsigned int f = *from++;
      switch (f)
	{
	case '+':
	  f = ' ';
	  break;
	case ';':			/* Potrebujeme rozeznavat, jedna-li se o jeden ze znaku, 	*/
	case '/':			/* ktere maji v URL specialni vyznam, nebo o jeho escapovanou   */
	case '?':			/* variantu, kterou musime po prekodovani opet escapovat	*/
	case ':':			/* To resime pridanim backslashe pred specialni znaky.		*/
	case '@':
	case '=':
	case '&':
	case '\r':
	case '\n':
	  *dbp++ = '\\';
	  break;
	case '%':
	  xbuf[state++] = f;
	percent:
	  while (state < 3)
	    {
	      if (from >= se)
		{
		  xbuf[state] = 0;
		  goto ibe;
		}
	      f = *from++;
	      xbuf[state++] = f;
	    }
	  state = 0;
	  f = (dehex(xbuf[1]) << 4) | dehex(xbuf[2]);
	  xbuf[0] = 0;
	  if (f == '\\')
	    *dbp++ = '\\';
	  break;
	}
      *dbp++ = f;
      if (dbp >= de - 1)
	{
	  *src = from;
	  *dst = dbp;
	  return F_OBF;
	}
    }
ibe:
  *src = from;
  *dst = dbp;
  return ((dbp >= de - 1) ? F_OBF : 0) | F_IBE;
}

/*
 * Filter handler pro cteni vstupu pres puvodni get_client_block().
 */
extern long __get_client_block(request_rec *, char *, int);

static int filter_get_input(struct filter *f, char **src, char *se, char **dst, char *de)
{
  int cnt = __get_client_block(f->request, *dst, de - *dst);

  if (cnt < 0)
    return F_ERR;
  if (!cnt)
    return F_IBE;
  *dst += cnt;
  return F_OBF;
}

/*
 * Filter handler pro zapis do uzivatelem dodaneho bufferu.
 */
static int filter_buf_output(struct filter *f, char **src, char *se, char **dst, char *de)
{
  return (se == f->buf_end) ? F_OBF : F_IBE;
}

/*
 * Filter handlery pro vstup ze stringu;
 */
static int filter_string_input(struct filter *f, char **src, char *se, char **dst, char *de)
{
  int len = f->buf_end - *src;

  if (len > de - *dst)
    len = de - *dst;
  if (len)
    {
      memcpy(*dst, *src, len);
      *dst += len;
      *src += len;
    }
  return ((*src == f->buf_end) ? F_IBE : 0) | ((*dst == de) ? F_OBF : 0);
}

/*
 * Filter handler pro prekodovavani znaku.
 */
struct recode_filter {
  struct filter f;
  struct conv_context cc;
};

static int filter_recode(struct filter *f, char **src, char *se, char **dst, char *de)
{
  struct recode_filter *F = (void *) f;
  struct conv_context *cc = &F->cc;
  int flags;

  cc->source = *src;
  cc->source_end = se;
  cc->dest = *dst;
  cc->dest_end = de;
  flags = conv_run(cc);
  *src = (char *) cc->source;
  *dst = cc->dest;
  return ((flags & CONV_DEST_END) ? F_OBF : 0) |
    ((flags & CONV_SOURCE_END) ? F_IBE : 0);
}

/*
 * Pokud nekdo zavola get_client_block pro ziskani dat poslanych
 * jako telo requestu, my je pripadne prekodujeme.
 */

static void setup_input_filters(request_rec *r, int esc, int src)
{
  struct recode_filter *rf;

  if (esc)
    add_filter(r, filter_deescape, 1024, sizeof(struct deescape_filter));
  if (src != CZ->form_cp)
    {
      rf = (void *) add_filter(r, filter_recode, IOBUFSIZE, sizeof(struct recode_filter));
      conv_set_charset(&rf->cc,
		       charset_info[src].conv_code,
		       charset_info[CZ->form_cp].conv_code);
    }
  if (esc)
    add_filter(r, filter_enescape, 1024, sizeof(struct enescape_filter));
}

long get_client_block(request_rec *r, char *buf, int size)
{
  int res;
  struct filter *f;

  if (!CZ->form_source || CZ->form_source == CZ->form_cp)
    return __get_client_block(r, buf, size);

  if (!CZ->first_filter)
    {
      add_filter(r, filter_get_input, 0, 0);
      setup_input_filters(r, CZ->is_form, CZ->form_source);
      add_filter(r, filter_buf_output, 0, 0);
    }

  f = CZ->last_filter;
  set_filter_buffer(f, buf, size);
  res = run_filter_chain(r);
  return (res < 0) ? res : (f->buf_mark - buf);
}

/*
 * Prekodovani argumentu uvedenych v URL.
 */
static void recode_args(request_rec *r)
{
  char *args = r->args;
  int ilen = strlen(args);
  int len = ilen + 16;
  char *buf, *buf2;
  struct filter *in, *out;

  in = add_filter(r, filter_string_input, 0, 0);
  set_filter_buffer(in, args, ilen);
  setup_input_filters(r, 1, CZ->czcode);
  out = add_filter(r, filter_buf_output, 0, 0);
  buf = alloca(len);
  set_filter_buffer(out, buf, len-1);
  for(;;)
    {
      int status = run_filter_chain(r);
      if (status < 0)
	return;
      if (out->finished)
	break;
      buf2 = buf;
      buf = alloca(2*len);
      memcpy(buf, buf2, len);
      set_filter_buffer(out, buf, 2*len-1);
      out->buf_mark += len;
      len += len;
    }
  *out->buf_mark = 0;
  r->args = pstrdup(r->pool, buf);
  reset_filters(r);
}

/*
 * Paklize jsme dostali nejake argumenty v URL, prekodujeme je.
 * Taktez ma-li request nejake telo, zjistime, jak je kodovane.
 * A nakonec jeste pridame environmentni promenne pro CGI scripty.
 */
static void prepare_input_encoding(request_rec *r)
{
  czech_dir_config *cfg = CZ_DIR_CONFIG(r);
  int have_body = table_get(r->headers_in, "Transfer-Encoding") ||
                  table_get(r->headers_in, "Content-Length");

  if (cfg->url_encode && r->args && CZ->czcode)	/* Argumenty */
    recode_args(r);

  if (cfg->form_encode && have_body)	/* Kodovani tela */
    {
      char *ct = table_get(r->headers_in, "Content-type");
      char *cts;
      int c, is_text;

      if (!ct)				/* Nemelo by nastat */
	ct = "x-unknown";
      for(cts=ct; *cts && *cts != ' ' && *cts != ';'; cts++)
	;
      c = *cts;
      *cts = 0;
      CZ->is_form = !strcasecmp(ct, "application/x-www-form-urlencoded");
      is_text = IsContentTypeForEncoding(r, ct);
      *cts = c;
      /* Mame prekodovat a vime, z ceho */
      if (CZ->is_form || is_text)
	{
	  CZ->form_source = guess_post_charset(r, ct); /* Mimo jine odstrani charset z content-typu */
	  if (CZ->form_cp != CS_ASCII && CZ->form_cp != CS_ISO_8859_1)
	    table_set(r->headers_in, "Content-type",
		      pstrcat(r->pool, ct, "; charset=", mime_charset(CZ->form_cp), NULL));
	  table_set(r->subprocess_env, "POST_CHARSET", mime_charset(CZ->form_source));
	}
      else
	CZ->form_source = 0;
    }
  else
    CZ->form_source = 0;

  table_set(r->subprocess_env, "URL_CHARSET", mime_charset(CZ->czcode));
  table_set(r->subprocess_env, "CLIENT_CHARSET", mime_charset(CZ->czcode));
  table_set(r->subprocess_env, "INPUT_CHARSET", mime_charset(CZ->form_cp));
  table_set(r->subprocess_env, "OUTPUT_CHARSET", mime_charset(CZ->czsource));

  /*
   * Promenne pro kompatibilitu s minulymi verzemi. Grrr.
   */
#ifdef CZECH_HACKS
  table_set(r->subprocess_env, "CZECH_OUTPUT_CP", mime_charset(CZ->czcode));
  table_set(r->subprocess_env, "CZECH_INPUT_CP", mime_charset(CZ->form_cp));
#endif
}

/* ----------------------------------------------------------- */
/* ---------------- CZECH TRANSLATE -------------------------- */
/* ----------------------------------------------------------- */

/* Zde se nastavi:
 * czcode      CP, v ktere se ma vysledek predat klientovi (UNKNOWN,
 *	       pokud to ma byt default -- nemame totiz jeste k dispozici
 *	       spravnou per-directory konfiguraci, pouze konfiguraci
 *	       od root directory prislusneho virtualniho serveru).
 * add_prefix  zda do vybranych URL ("HREF","Location:" apod.) pridavat
 *             kodovaci prefix (ke konfiguraci prihledneme v czech_mime()).
 * Jestli je modul vypnuty, jsou vsechny kodovaci promenne nastavene
 * na 0 (= nekodovat).
 * Czsource a form_cp se nastavuji az v czech_mime().
 */

/*-------------------- czech_translate --------------------- */

static int czech_translate(request_rec *r)
{
  czech_dir_config *cfg = CZ_DIR_CONFIG(r);

  DBG(("\n========== translate: '%s' =========\n", r->uri));

  r->mcz_cfg = pcalloc(r->pool, sizeof(mcz_config));

  /*
   * Pozor, toto je globalni nastaveni per-server, pripadne <Location>,
   * jelikoz .htaccess se dosud nehledal. Lokalni zakazani mod_czech (tedy
   * takove, ktere povoluje zpracovani prefixu a postfixu, ale uz ne
   * prekodovavani, se resi v czech_mime(). Kazdopadne neni-li mod_czech
   * povolen, skoncime s czcode = form_cp = add_prefix = 0.
   */
  if (!cfg->encode)
      return(DECLINED);

  CZ->form_source = UNDEFINED;

  switch (AnalyzeRequest(r, r->uri, 1)) 
  {
    case 0: /* Nic jsem v requestu nenasel */
            if(r->prev) {           /* internal redirect*/
	        CZ->czcode     = prevCZ->czcode;
		CZ->add_prefix = prevCZ->add_prefix;
		CZ->form_source = prevCZ->form_source;
	    }
	    else
	    if(r->main) {           /* subrequest */
	        CZ->czcode     = mainCZ->czcode;
		CZ->add_prefix = mainCZ->add_prefix;
		CZ->form_source = mainCZ->form_source;
	    }
	    else
	        CZ->czcode = CS_ASCII;    /* default, pokud neuspeje guess */
	    break;

    case 1: /* V requestu jsem nasel explicitni CP ve forme prefixu */ 
 	    CZ->add_prefix = 1; 
	    break;

    case 2: /* redirekt: prisel postfix, nebo nestandardni prefix */
 	    CZ->add_prefix = 1;
	    /* Nestandartni prefix nebo postfix nam v subrequestech nevadi.
	     * Problemy pak maji jen _spatne_ redirekce z cgi-skriptu
	     * (tj. "Location: dokument.html.us-ascii", pokud bylo mineno
	     * jako interni redirekce). Povolit redirekce vzdy neni mozne,
	     * pak by zas nefungovaly SSI.
	     */
	    if (!r->prev && !r->main) {
	        table_set(r->headers_out, "Location", 
			  construct_url(r->pool, r->uri, r->server));
		return HTTP_MOVED_PERMANENTLY;
	    }
	    break;
  }

  DBG(("(translate end): Czcode=%d, czsource=%d\n",CZ->czcode,CZ->czsource));

  return(DECLINED);
}


/* ----------------------------------------------------------- */
/* ---------------- CZECH MIME CHECKER ----------------------- */
/* ----------------------------------------------------------- */

/* Pri sub_req_lookup_file() se czech_translate() nevola, proto jsou vsechna
 * nastaveni, ktera primo nesouvisi s URL, provedena zde.
 * Zde se nastavi:
 *      czcode      CP pro clienta (pokud ji nenastavil czech_translate())
 *      czsource    CP zdrojoveho dokumentu
 *      form_cp     CP, do ktere jsou prekodovana data z formularu
 * Oboje je urceno v konf. adresare nebo serveru. Czsource lze jeste upravit
 * primo extenzi' servirovaneho souboru v mod_mime.
 */

/*-------------------- czech_mime --------------------- */
static int czech_mime (request_rec *r)
{
  czech_dir_config *cfg = CZ_DIR_CONFIG(r);

  DBG(("========== mime: =========\n"));
  DBG(("URI='%s', file='%s'\n",r->uri,r->filename));

 /* Pri sub_req_lookup_file(): */
  if(!r->mcz_cfg)
  {
      r->mcz_cfg = pcalloc(r->pool, sizeof(mcz_config));
      CZ->czcode      = mainCZ->czcode;       /* (r->main je def.) */
      CZ->add_prefix  = mainCZ->add_prefix;
      CZ->form_source = UNDEFINED;
  }

 /* Vse nasledujici lze zakazat direktivou "CzechModule Off": */
  if (!cfg->encode)
    return(DECLINED);

  if(r->prev) {
    CZ->master_cf = prevCZ->master_cf;
  } else if(r->main) {
    CZ->master_cf = mainCZ->master_cf;
    /* Oprava chyby Apache. Jinak by dobre nefungovalo http/1.1 cachovani. */
    r->proto_num  = r->main->proto_num;
  } else {
    struct tag_buff *tb = palloc(r->pool, sizeof(struct tag_buff));
    CZ->master_cf = CZ;
    CZ->master_cf->mycharset = UNDEFINED;
    CZ->master_cf->tag = tb;
    tb->pos = NULL;
    tb->state = 0;
  }

  CZ->czsource = MERGE_INTS(CS_ISO_8859_2, cfg->czsource);
  CZ->form_cp  = MERGE_INTS(CS_ISO_8859_2, cfg->form_cp);

  CZ->tag = CZ->master_cf->tag;

  DBG(("(czech_mime end): Czcode=%d, czsource=%d\n",CZ->czcode,CZ->czsource));

  return (DECLINED);
}


/* ----------------------------------------------------------- */
/* ---------------- CZECH FIXUP ------------------------------ */
/* ----------------------------------------------------------- */

/*-------------------- czech_fixup --------------------- */
static int czech_fixup (request_rec *r)
{
  czech_dir_config *cfg = CZ_DIR_CONFIG(r);

  DBG(("========== fixups =========\n"));
  DBG(("(fixups beg): Czcode=%d, czsource=%d\n",CZ->czcode,CZ->czsource));
  DBG(("URI='%s', file='%s'\n",r->uri,r->filename));

 /* Vse nasledujici lze zakazat direktivou "CzechModule Off": */
  if(!cfg->encode)
  { 
      CZ->czcode      = 0;
      CZ->add_prefix  = 0;
      CZ->form_source = 0;
      return(DECLINED);
  }

  /* Kodovani zvolene pres <!--mycharset=xxx--> respektujeme i v includech */
  if (CZ->master_cf->mycharset != UNDEFINED)
    CZ->czsource = CZ->master_cf->mycharset;

  /* Dame dohromady CP vysledku, pokud nebyla urcena explicitne */
  if (!CZ->add_prefix  && !r->main  && !r->prev)
  {
    int code = guess_charset(r);
    DBG(("Guessed %s\n", (code >= 0) ? mime_charset(code) : "UNKNOWN"));
    DFLUSH;
    if (code >= 0)
	CZ->czcode = code;
    else if (cfg->guess_charset  && (cfg->fallback_charset >= 0))
        CZ->czcode = cfg->fallback_charset;
    else
    {
	if (IsContentTypeForEncoding(r,r->content_type)  ||
	    r->content_type  && !strcasecmp(r->content_type, DIR_MAGIC_TYPE))
	{
	    table_set(r->headers_out, "Location",
		      construct_url(r->pool,
				    pstrcat(r->pool, "/chcp", r->uri, NULL),
				    r->server));
	    return HTTP_MOVED_PERMANENTLY;
	}
	CZ->czcode = 0;
    }
  }

  if(!CZ->czsource) CZ->czcode = 0;

  /* Globalni zakaz kodovacich prefixu: */
  if(!cfg->prepend_prefix)
      CZ->add_prefix = 0;

  /* Zjistime zdrojovou znakovou sadu pro data zasilana serverem */
  prepare_input_encoding(r);

  return (DECLINED);
}


/* ----------------------------------------------------- */
/* ------- GENEROVANI STRANKY CHCP - ZMENA CP ---------- */
/* ----------------------------------------------------- */

/* ---------------------- GetRootPos ------------------------- */
static int GetRootPos(char *ref)
{
 char *pos;

 pos = strstr(ref,"://");
 if(!pos) return(0);
 pos+=3;
 while((*pos != 0) && (*pos != '/')) pos++;
 if(!*pos) return(0);
 return(pos-ref);
}

/* ------------------- SendCPChoice ------------------------- */
static int SendCPChoice (request_rec *r)
{
int n, i;
char buf[1000];
char *href;
czech_server_config *sconf = CZ_CONFIG(r);
czech_dir_config *cfg = CZ_DIR_CONFIG(r);

  if (!cfg->encode)
    return HTTP_NOT_FOUND;

  CZ->add_prefix = 1;		/* Abychom nedostali Vary: a spol. */
  CZ->czcode = CS_ASCII;
  r->content_type = "text/html";

  soft_timeout ("send", r);
  send_http_header (r);

  init_conv_ctxt (r, CS_ISO_8859_2, CS_ASCII);
  CZ->add_prefix = 0;
  EncodeStream(r, choice1, strlen(choice1));

  /* Pokud nejsme zavolani s cestou k dokumentu, zkusime pouzit Referer: */
  if ((!r->path_info  || !r->path_info[0])  &&
      (href = table_get(r->headers_in, "Referer")))
    {
      href = pstrdup(r->pool, href);
      n = GetRootPos(href); 
      AnalyzeRequest(r, &href[n], 0);
    }
  else
    {
      char port[16];
      if (r->server->port != 80)
	sprintf(port, ":%d", r->server->port);
      else
	port[0] = 0;
      href = pstrcat(r->pool, "http://", r->server->server_hostname, port,
	        (r->path_info && r->path_info[0] ? r->path_info : "/"), NULL);
    }

  CZ->add_prefix = 1;
  for(i=0; i<CS_MAX; i++)
    if (charset_info[i].comment && GetDefaultPrefix(sconf, i))
      {
	CZ->czcode = i;
	init_conv_ctxt (r, CS_ISO_8859_2, i);
	n = sprintf(buf, choice_one, AddPrefixToURL(r, href), charset_info[i].comment);
	EncodeStream(r, buf, n);
      }

  CZ->add_prefix = 0;
  init_conv_ctxt (r, CS_ISO_8859_2, CS_ASCII);
  EncodeStream(r, choice2, strlen(choice2));

  kill_timeout(r);

  return OK;
}


/* ------------------------------------------------------- */
/* --------------  SEKCE NACITANI KONFIGURACI ------------ */
/* ------------------------------------------------------- */

/* -------------- create_czech_server_config --------------------- */
static void *create_czech_server_config(pool *p, server_rec *s)
{
  czech_server_config *sconf = palloc(p, sizeof(czech_server_config));

  sconf->czprefixes = make_table(p, 20);
  sconf->charset_aliases = make_table(p, 20);
  sconf->postfix_aliases = make_table(p, 20);
  return sconf;
}

/* -------------- merge_czech_server_config --------------------- */
static void *merge_czech_server_config(pool *p, void *basev, void *addv)
{
  czech_server_config *add = addv,
                      *base = basev,
                      *new = palloc(p, sizeof(czech_server_config));

  /*
   * Toto je ponekud neprirozene poradi, nicmene je nutne, jelikoz nam nejde
   * o override, ale naopak o doplneni do tabulky se zachovanim puvodnich
   * defaultu.
   */
  new->czprefixes = overlay_tables(p, base->czprefixes, add->czprefixes);
  new->charset_aliases = overlay_tables(p, base->charset_aliases, add->charset_aliases);
  new->postfix_aliases = overlay_tables(p, base->postfix_aliases, add->postfix_aliases);
  return new;
}

/* -------------- create_czech_dir_config --------------------- */
static void *create_czech_dir_config(pool *p, char *dummy)
{
  czech_dir_config *cfg = palloc(p, sizeof(czech_dir_config));

  cfg->encode = UNDEFINED;
  cfg->form_encode = UNDEFINED;
  cfg->url_encode = UNDEFINED;
  cfg->czsource = UNDEFINED;
  cfg->form_cp = UNDEFINED;
  cfg->prepend_prefix = UNDEFINED;
  cfg->do_cache = UNDEFINED;
  cfg->set_lm = UNDEFINED;
  cfg->add_cp_to_ct = UNDEFINED;
  cfg->allow_dyn_cp_changes = UNDEFINED;
  cfg->fallback_charset = UNDEFINED;
  cfg->guess_charset = UNDEFINED;
  cfg->content_types = make_table(p, 8);
  cfg->czech_servers = make_table(p, 8);
  cfg->src_charsets  = make_table(p, 20);
  cfg->allow_html_parsing = UNDEFINED;

  return cfg;
}

/* -------------- merge_czech_dir_config --------------------- */
static void *merge_czech_dir_config(pool *p, void *basev, void *overv)
{
  czech_dir_config *old = basev,
                   *over = overv,
                   *new = palloc(p, sizeof(czech_dir_config));

  new->encode = MERGE_INTS(old->encode, over->encode);
  new->form_encode = MERGE_INTS(old->form_encode, over->form_encode);
  new->url_encode = MERGE_INTS(old->url_encode, over->url_encode);
  new->czsource = MERGE_INTS(old->czsource, over->czsource);
  new->form_cp = MERGE_INTS(old->form_cp, over->form_cp);
  new->prepend_prefix = MERGE_INTS(old->prepend_prefix, over->prepend_prefix);
  new->do_cache = MERGE_INTS(old->do_cache, over->do_cache);
  new->set_lm = MERGE_INTS(old->set_lm, over->set_lm);
  new->add_cp_to_ct = MERGE_INTS(old->add_cp_to_ct, over->add_cp_to_ct);
  new->allow_dyn_cp_changes = MERGE_INTS(old->allow_dyn_cp_changes, over->allow_dyn_cp_changes);
  new->fallback_charset = MERGE_INTS(old->fallback_charset, over->fallback_charset);
  new->guess_charset = MERGE_INTS(old->guess_charset, over->guess_charset);
  new->content_types = overlay_tables(p, over->content_types, old->content_types);
  new->czech_servers = overlay_tables(p, over->czech_servers, old->czech_servers);
  new->src_charsets = overlay_tables(p, over->src_charsets, old->src_charsets);
  new->allow_html_parsing = MERGE_INTS(old->allow_html_parsing, over->allow_html_parsing);

  return new;
}

/*
 * Zpracovani booleovskych konfiguracnich optionu.
 */

/* -------------- SET_BOOLEAN a ruzne direktivy --------------------- */
#define SET_BOOLEAN(func, var)						\
static const char * func (cmd_parms *cmd, czech_dir_config *cfg, int flag) \
{									\
  cfg->var = flag;						 	\
  return NULL;								\
}

SET_BOOLEAN(set_czech, encode)
SET_BOOLEAN(set_prefixes, prepend_prefix)
SET_BOOLEAN(set_do_cache, do_cache)
SET_BOOLEAN(set_last_mod, set_lm)
SET_BOOLEAN(set_add_cp_to_ct, add_cp_to_ct)
SET_BOOLEAN(set_form_encode, form_encode)
SET_BOOLEAN(set_url_encode, url_encode)
SET_BOOLEAN(set_dyn_cp, allow_dyn_cp_changes)
SET_BOOLEAN(set_guess_charset, guess_charset)
SET_BOOLEAN(set_html_parsing, allow_html_parsing)

/* -------------- set_source_cp --------------------- */
static const char *set_source_cp(cmd_parms *cmd, czech_dir_config *cfg, char *cp)
{
  int cpi;

  if ((cpi = config_charset(CZ_CONFIG(cmd), cp)) < 0)
    return "Unknown name of character set";
  cfg->czsource = cpi;
  return NULL;
}

/* -------------- set_form_cp --------------------- */
static const char *set_form_cp(cmd_parms *cmd, czech_dir_config *cfg, char *cp)
{
  int cpi;

  if ((cpi = config_charset(CZ_CONFIG(cmd), cp)) < 0)
    return "Unknown name of character set";
  cfg->form_cp = cpi;
  return NULL;
}

/* -------------- add_content_type --------------------- */
static const char *add_content_type(cmd_parms *cmd, czech_dir_config *cfg, char *ct)
{
  table_set(cfg->content_types, ct, ct);
  return NULL;
}

/* ---------------- add_src_charset --------------- */
static const char *add_src_charset(cmd_parms *cmd, czech_dir_config *cfg, 
			    char *cp, char *ext)
{
  if (config_charset(CZ_CONFIG(cmd), cp) < 0)
    return "Unknown name of character set";
  if (*ext == '.') ++ext;
  TableSet(cfg->src_charsets, ext, cp);
  return NULL;
}

/* -------------- add_czech_server --------------------- */
static const char *add_czech_server(cmd_parms *cmd, czech_dir_config *cfg, const char *args)
{
  array_header *srvports;
  char *srvname, *p;
  int port = -1;

  srvname = getword_conf (cmd->pool, &args);

  if (*srvname == '\0')
      return pstrcat(cmd->pool, 
		  "AddModCzechServer requires at least one argument", NULL);

  if (p = strchr(srvname, ':')) {
      srvname = pstrndup(cmd->pool, srvname, p - srvname);
      p++;
  }
  else {
      srvname = pstrdup(cmd->pool, srvname);      /* copy it into my pool */
  }

  if (!(srvports = (array_header *) table_get(cfg->czech_servers, srvname)))
      srvports = make_array(cmd->pool, 5, sizeof(int));

  do {
      if (p == NULL)  continue;
      if (isdigit(*p)) {
	  port = atoi(p);
	  *((int *) push_array(srvports)) = port;
      }
      else
	  return pstrcat(cmd->pool,
	 "Other arguments for AddModCzechServer must be port numbers", NULL);
  } while (*(p = getword_conf(cmd->temp_pool, &args)) != '\0');

  table_set_element (cfg->czech_servers, srvname, (char *) srvports);
  return NULL;
}

/* -------------- add_postfix_alias --------------------- */
static const char *add_postfix_alias(cmd_parms *cmd, void *dummy, char *std, char *new)
{
  czech_server_config *sconf = CZ_CONFIG(cmd);

  if (!TableGet(sconf->czprefixes, std))
    return "Unknown prefix";
  TableSet(sconf->postfix_aliases, new, std);
  return NULL;
}

/* -------------- add_prefix --------------------- */
static const char *add_prefix(cmd_parms *cmd, void *dummy, char *std, char *new)
{
  czech_server_config *cfg = CZ_CONFIG(cmd);
  int i = config_charset(cfg, std);

  if (i < 0)
    return "Unknown name of character set";
  TableSet(cfg->czprefixes, new, mime_charset(i));
  return NULL;
}

/* -------------- add_alias ---------------------- */
static const char *add_alias(cmd_parms *cmd, void *dummy, char *std, char *new)
{
  czech_server_config *cfg = CZ_CONFIG(cmd);
  int i = config_charset(cfg, std);

  if (i < 0)
    return "Unknown name of character set";
  TableSet(cfg->charset_aliases, new, mime_charset(i));
  return NULL;
}

static const char *
set_fallback_charset(cmd_parms *cmd, czech_dir_config *cfg, char *arg)
{
  int cs;

  if (!strcasecmp(arg, "CHCP"))
    cfg->fallback_charset = -1;
  else if ((cs = config_charset(CZ_CONFIG(cmd), arg)) >= 0)
    cfg->fallback_charset = cs;
  else
    return "FallbackCharset: invalid argument";
  return NULL;
}

/* -------------- Command declarations -------------------- */
static command_rec czech_cmds[] = {

  /* Per-server configuration */
  { "AddCharsetPrefix", add_prefix, NULL, RSRC_CONF, ITERATE2, NULL },
  { "AddCharsetAlias", add_alias, NULL, RSRC_CONF, ITERATE2, NULL },
  { "AddPostfix", add_postfix_alias, NULL, RSRC_CONF, ITERATE2, NULL },

  /* Per-directory configuration */
  { "CzechModule", set_czech, NULL, OR_FILEINFO, FLAG, NULL },
  { "SourceCharset", set_source_cp, NULL, OR_FILEINFO, TAKE1, NULL },
  { "PrependPrefixes", set_prefixes, NULL, OR_FILEINFO, FLAG, NULL },
  { "EncodeContentType", add_content_type, NULL, OR_FILEINFO, TAKE1, NULL },
  { "CacheEncodedDocs", set_do_cache, NULL, OR_FILEINFO, FLAG, NULL },
  { "SetLastModified", set_last_mod, NULL, OR_FILEINFO, FLAG, NULL },
  { "AddCharsetToContentType", set_add_cp_to_ct, NULL, OR_FILEINFO, FLAG, NULL },
  { "FormURLEncoding", set_url_encode, NULL, OR_FILEINFO, FLAG, NULL },
  { "FormDataEncoding", set_form_encode, NULL, OR_FILEINFO, FLAG, NULL },
  { "FormDataCharset", set_form_cp, NULL, OR_FILEINFO, TAKE1, NULL },
  { "AllowMETASourceCharset", set_dyn_cp, NULL, OR_FILEINFO, FLAG, NULL },
  { "FallbackCharset", set_fallback_charset, NULL, OR_FILEINFO, TAKE1, NULL },
  { "GuessCharset", set_guess_charset, NULL, OR_FILEINFO, FLAG, NULL },
  { "AddModCzechServer", add_czech_server, NULL, OR_FILEINFO, RAW_ARGS, NULL },
  { "AddSourceCharset", add_src_charset, NULL, OR_FILEINFO,ITERATE2, NULL },
  { "AllowHTMLParsing", set_html_parsing, NULL, OR_FILEINFO, FLAG, NULL },

  /* Compatibility glue for .htaccess */
  { "SourceCzechCodePageLocal", set_source_cp, NULL, OR_FILEINFO, TAKE1, NULL },
  { "FormDataCPTargetLocal", set_form_cp, NULL, OR_FILEINFO, TAKE1, NULL },

  { NULL },
};

/* -------------- Handlers declarations ------------------- */
static handler_rec czech_handlers[] = {
{ "cp-choice", SendCPChoice },
{ NULL }
};

/* --------------- Module definition ----------------------- */
module czech_module = {
   STANDARD_MODULE_STUFF,
   NULL,			/* initializer */
   create_czech_dir_config,	/* dir config creater */
   merge_czech_dir_config,	/* dir merger --- default is to override */
   create_czech_server_config,	/* server config */
   merge_czech_server_config,	/* merge server config */
   czech_cmds,			/* command table */
   czech_handlers,		/* handlers */
   czech_translate,		/* filename translation */
   NULL,			/* check_user_id */
   NULL,			/* check auth */
   NULL,			/* check access */
   czech_mime,			/* type_checker */
   czech_fixup,			/* fixups */
   NULL				/* logger */
};
